# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Functions and beam DoFn for reading and encoding images."""

import base64
import logging
import os
from typing import Any, Dict, Generator, Tuple

import apache_beam as beam
from apache_beam.metrics import Metrics
import tensorflow as tf
from PIL import Image


BASE64_ALTCHARS = b'-_'


def mode_to_channel(mode: str) -> int:
  """Returns number of channels depending on PIL image `mode`s."""

  return 1 if 'L' in mode else 3


def channel_to_mode(channels: int) -> str:
  """Returns PIL image mode depending on number of `channels`."""

  return 'L' if channels == 1 else 'RGB'


def encode(image: Image):
  """Returns base64-encoded image data.

  Args:
    image: PIL image.
  """

  return base64.b64encode(image.tobytes(), altchars=BASE64_ALTCHARS)


def decode(b64_bytes, width, height, channels) -> Image:
  """Decodes an image from base64-encoded data."""

  image_bytes = base64.b64decode(b64_bytes, altchars=BASE64_ALTCHARS)
  mode = channel_to_mode(channels)
  return Image.frombytes(mode, (width, height), image_bytes)


def load(image_uri):
  """Loads an image using Pillow.

  Supported formats:
  https://pillow.readthedocs.io/en/5.1.x/handbook/image-file-formats.html
  """

  try:
    with tf.io.gfile.GFile(image_uri, 'rb') as f:
      return Image.open(f)
  except tf.python.framework.errors_impl.NotFoundError as e:
    raise OSError('File {} was not found.'.format(image_uri)) from e


# pylint: disable=abstract-method

class ExtractImagesDoFn(beam.DoFn):
  """Adds image to PCollection."""

  def __init__(self, image_uri_key: str):
    """Constructor."""
    super().__init__()
    self.image_uri_key = image_uri_key
    self.image_good_counter = Metrics.counter(self.__class__, 'image_good')
    self.image_bad_counter = Metrics.counter(self.__class__, 'image_bad')


  # pylint: disable=unused-argument
  def process(
      self,
      element: Dict[str, Any],
      *args: Tuple[Any, ...],
      **kwargs: Dict) -> Generator[Dict[str, Any], None, None]:
    """Loads image and creates image features.

    This DoFn extracts an image being stored on local disk or GCS and
    yields a base64 encoded image, the image height, image width, and channels.
    """
    d = {}

    try:
      image_uri = element.pop(self.image_uri_key)
      image = load(image_uri)
      element['image_name'] = os.path.split(image_uri)[-1]
      d['image'] = encode(image)
      d['image_width'], d['image_height'] = image.size
      d['image_channels'] = mode_to_channel(image.mode)
      self.image_good_counter.inc()

    # pylint: disable=broad-except
    except Exception as e:
      logging.warning('Could not load image: %s', image_uri)
      logging.error('Exception was: %s', str(e))
      self.image_bad_counter.inc()
      d['split'] = 'DISCARD'

    element.update(d)
    yield element



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for `dataset.py`."""

import os
import tempfile
import unittest

from tfrecorder import dataset
from tfrecorder import schema
from tfrecorder import test_utils

# pylint: disable=protected-access

class ValidateTFRecordDirTest(unittest.TestCase):
  """Tests `_validate_tfrecord_dir`."""

  def setUp(self):
    self._temp_dir_obj = tempfile.TemporaryDirectory()
    self.temp_dir = self._temp_dir_obj.name

  def tearDown(self):
    self._temp_dir_obj.cleanup()

  def test_ok(self):
    """Checks that function works as expected when TFT dirs are present."""
    os.makedirs(os.path.join(self.temp_dir, dataset.TRANSFORMED_METADATA_DIR))
    os.makedirs(os.path.join(self.temp_dir, dataset.TRANSFORM_FN_DIR))
    dataset._validate_tfrecord_dir(self.temp_dir)

  def test_missing_metadata_dir(self):
    """Check exception raised when metadata directory missing."""

    with self.assertRaises(FileNotFoundError):
      os.makedirs(os.path.join(self.temp_dir, dataset.TRANSFORM_FN_DIR))
      dataset._validate_tfrecord_dir(self.temp_dir)

  def test_missing_transform_fn_dir(self):
    """Check exception raised when transform_fn directory missing."""
    with self.assertRaises(FileNotFoundError):
      os.makedirs(
          os.path.join(self.temp_dir, dataset.TRANSFORMED_METADATA_DIR))
      dataset._validate_tfrecord_dir(self.temp_dir)

  def test_missing_tf_transform_dirs(self):
    """Check exception raised when both TFT transform directories missing."""
    with self.assertRaises(FileNotFoundError):
      dataset._validate_tfrecord_dir(self.temp_dir)


class LoadTest(unittest.TestCase):
  """Tests `load` function."""

  def setUp(self):
    self.tfrecord_dir = test_utils.TEST_TFRECORDS_DIR

  def test_load_all_splits(self):
    """Test case where all TFRecord splits can be loaded."""
    dataset_dict = dataset.load(self.tfrecord_dir)
    self.assertEqual(len(dataset_dict), 3)
    self.assertCountEqual(
        list(dataset_dict.keys()), schema.allowed_split_values[:-1])


if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for common."""

import os
import unittest
import tempfile

from tfrecorder import common
from tfrecorder import constants


class CopyLogTest(unittest.TestCase):
  """Misc tests for _copy_logfile_to_gcs."""

  def test_valid_copy(self):
    """Test valid file copy."""
    with tempfile.TemporaryDirectory() as tmpdirname:
      text = 'log test log test'
      infile = os.path.join(tmpdirname, 'foo.log')
      with open(infile, 'w') as f:
        f.write(text)
      common.copy_logfile_to_gcs(infile, tmpdirname)

      outfile = os.path.join(tmpdirname, constants.LOGFILE)
      with open(outfile, 'r') as f:
        data = f.read()
        self.assertEqual(text, data)

  def test_invalid_copy(self):
    """Test invalid file copy."""
    with tempfile.TemporaryDirectory() as tmpdirname:
      infile = os.path.join(tmpdirname, 'foo.txt')
      with self.assertRaises(FileNotFoundError):
        common.copy_logfile_to_gcs(infile, tmpdirname)

# pylint: disable=protected-access
if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a pandas DataFrame accessor for TFRecorder.

accessor.py contains TFRecorderAccessor which provides a pandas DataFrame
accessor.  This accessor allows us to inject the to_tfr() function into
pandas DataFrames.
"""

from typing import Any, Dict, Optional, Union
import pandas as pd
from IPython.core import display

from tfrecorder import client
from tfrecorder import constants
from tfrecorder import schema


@pd.api.extensions.register_dataframe_accessor('tensorflow')
class TFRecorderAccessor:
  """DataFrame Accessor class for TFRecorder."""

  def __init__(self, pandas_obj):
    self._df = pandas_obj

  # pylint: disable=too-many-arguments
  def to_tfr(
      self,
      output_dir: str,
      schema_map: Dict[str, schema.SchemaMap] = schema.image_csv_schema,
      runner: str = 'DirectRunner',
      project: Optional[str] = None,
      region: Optional[str] = None,
      tfrecorder_wheel: Optional[str] = None,
      dataflow_options: Union[Dict[str, Any], None] = None,
      job_label: str = 'to-tfr',
      compression: Optional[str] = 'gzip',
      num_shards: int = 0) -> Dict[str, Any]:
    """TFRecorder Pandas Accessor.

    TFRecorder provides an easy interface to create image-based tensorflow
    records from a dataframe containing GCS locations of the images and labels.

    Usage:
      import tfrecorder

      df.tfrecorder.to_tfr(
          output_dir='gcs://foo/bar/train',
          runner='DirectRunner',
          compression='gzip',
          num_shards=10)

    Args:
      schema_map: A dict mapping column names to supported types.
      output_dir: Local directory or GCS Location to save TFRecords to.
        Note: GCS required for DataflowRunner
      runner: Beam runner. Can be DirectRunner or  DataflowRunner.
      project: GCP project name (Required if DataflowRunner).
      region: GCP region name (Required if DataflowRunner).
      tfrecorder_wheel: Path to the tfrecorder wheel Dataflow will run.
        (create with 'python setup.py sdist' or
        'pip download tfrecorder --no-deps')
      dataflow_options: Optional dictionary containing Dataflow options.
      job_label: User supplied description for the beam job name.
      compression: Can be 'gzip' or None for no compression.
      num_shards: Number of shards to divide the TFRecords into. Default is
          0 = no sharding.
    Returns:
      job_results: A dictionary of job results.
    """
    display.display(
        display.HTML(
            '<b>Logging output to /tmp/{} </b>'.format(constants.LOGFILE)))

    r = client.create_tfrecords(
        self._df,
        output_dir=output_dir,
        schema_map=schema_map,
        runner=runner,
        project=project,
        region=region,
        tfrecorder_wheel=tfrecorder_wheel,
        dataflow_options=dataflow_options,
        job_label=job_label,
        compression=compression,
        num_shards=num_shards)
    return r



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFRecorder Beam Pipeline.

This file implements the full Beam pipeline for TFRecorder.
"""

import collections
from typing import Any, Callable, Dict, Generator, List, Optional, Union

import functools
import logging
import os

import apache_beam as beam
from apache_beam import pvalue
import pandas as pd
import tensorflow_transform as tft
from tensorflow_transform import beam as tft_beam

from tfrecorder import beam_image
from tfrecorder import common
from tfrecorder import schema
from tfrecorder import types


def _get_job_name(job_label: str = None) -> str:
  """Returns Beam runner job name.

  Args:
    job_label: A user defined string that helps define the job.

  Returns:
    A job name compatible with apache beam runners, including a time stamp to
      insure uniqueness.
  """

  job_name = 'tfrecorder-' + common.get_timestamp()
  if job_label:
    job_label = job_label.replace('_', '-')
    job_name += '-' + job_label

  return job_name


def _get_job_dir(output_path: str, job_name: str) -> str:
  """Returns Beam processing job directory."""

  return os.path.join(output_path, job_name)


def _get_pipeline_options(
    runner: str,
    job_name: str,
    job_dir: str,
    project: str,
    region: str,
    tfrecorder_wheel: str,
    dataflow_options: Union[Dict[str, Any], None]
    ) -> beam.pipeline.PipelineOptions:
  """Returns Beam pipeline options."""

  options_dict = {
      'runner': runner,
      'staging_location': os.path.join(job_dir, 'staging'),
      'temp_location': os.path.join(job_dir, 'tmp'),
      'job_name': job_name,
      'teardown_policy': 'TEARDOWN_ALWAYS',
      'save_main_session': True,
      'pipeline_type_check': False,
  }

  if project:
    options_dict['project'] = project
  if region:
    options_dict['region'] = region
  if runner == 'DataflowRunner':
    options_dict['extra_packages'] = [tfrecorder_wheel]
  if dataflow_options:
    options_dict.update(dataflow_options)

  return beam.pipeline.PipelineOptions(flags=[], **options_dict)


def _partition_fn(
    element: Dict[str, str],
    unused_num_partitions: int = -1,
    split_key: str = 'split') -> int:
  """Returns index used to partition an element from a PCollection."""
  del unused_num_partitions
  dataset_type = element[split_key].decode('utf-8')
  try:
    index = schema.SplitKeyType.allowed_values.index(dataset_type)
  except ValueError as e:
    logging.warning('Unable to index dataset type %s: %s.',
                    dataset_type, str(e))
    index = schema.SplitKeyType.allowed_values.index('DISCARD')
  return index

def _get_write_to_tfrecord(output_dir: str,
                           prefix: str,
                           compress: bool = True,
                           num_shards: int = 0) \
                           -> beam.io.tfrecordio.WriteToTFRecord:
  """Returns `beam.io.tfrecordio.WriteToTFRecord` object.
  This configures a Beam sink to output TFRecord files.
  Args:
    output_dir: Directory to output TFRecord files.
    prefix: TFRecord file prefix.
    compress: If True, GZip compress TFRecord files.
    num_shards: Number of file shards to split the TFRecord data.
  """

  path = os.path.join(output_dir, prefix)
  suffix = '.tfrecord'
  if compress:
    compression_type = 'gzip'
    suffix += '.gz'
  else:
    compression_type = 'uncompressed'

  return beam.io.tfrecordio.WriteToTFRecord(
      path,
      file_name_suffix=suffix,
      compression_type=compression_type,
      num_shards=num_shards,
  )


def _preprocessing_fn(inputs: Dict[str, Any],
                      schema_map: Dict[str, collections.namedtuple]):
  """TensorFlow Transform preprocessing function."""

  outputs = {}
  for name, supported_type in schema_map.items():
    if supported_type.type_name == 'string_label':
      outputs[name] = tft.compute_and_apply_vocabulary(inputs[name])
    else:
      outputs[name] = inputs[name]
  return outputs


# pylint: disable=abstract-method

class ToCSVRows(beam.DoFn):
  """Adds image to PCollection."""

  def __init__(self):
    """Constructor."""
    super().__init__()
    self.row_count = beam.metrics.Metrics.counter(self.__class__, 'row_count')


  # pylint: disable=unused-argument
  # pylint: disable=arguments-differ
  def process(
      self,
      element: List[str],
      ) -> Generator[Dict[str, Any], None, None]:
    """Converts an input pandas DataFrame row (List of strings) into a flat
      column seperated row. This is necessary so that the input DataFrame
      can operate with TF Transform Coders."""
    element = ','.join([str(item) for item in element])
    self.row_count.inc()
    yield element


def get_split_counts(df: pd.DataFrame, split_key: str):
  """Returns number of rows for each data split type given dataframe."""
  assert split_key in df.columns
  return df[split_key].value_counts().to_dict()


def _transform_and_write_tfr(
    dataset: pvalue.PCollection,
    tfr_writer: Callable[[], beam.io.tfrecordio.WriteToTFRecord],
    raw_metadata: types.BeamDatasetMetadata,
    preprocessing_fn: Optional[Callable] = None,
    transform_fn: Optional[types.TransformFn] = None,
    label: str = 'data'):
  """Applies TF Transform to dataset and outputs it as TFRecords."""

  dataset_metadata = (dataset, raw_metadata)

  if transform_fn:
    transformed_dataset, transformed_metadata = (
        (dataset_metadata, transform_fn)
        | f'Transform{label}' >> tft_beam.TransformDataset())
  else:
    if not preprocessing_fn:
      preprocessing_fn = lambda x: x
    (transformed_dataset, transformed_metadata), transform_fn = (
        dataset_metadata
        | f'AnalyzeAndTransform{label}' >>
        tft_beam.AnalyzeAndTransformDataset(preprocessing_fn))

  transformed_data_coder = tft.coders.ExampleProtoCoder(
      transformed_metadata.schema)
  _ = (
      transformed_dataset
      | f'Encode{label}' >> beam.Map(transformed_data_coder.encode)
      | f'Write{label}' >> tfr_writer(prefix=label.lower()))

  return transform_fn


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def build_pipeline(
    df: pd.DataFrame,
    job_label: str,
    runner: str,
    project: str,
    region: str,
    output_dir: str,
    compression: str,
    num_shards: int,
    schema_map: Dict[str, collections.namedtuple],
    tfrecorder_wheel: str,
    dataflow_options: Dict[str, Any]) -> beam.Pipeline:
  """Runs TFRecorder Beam Pipeline.

  Args:
    df: Pandas DataFrame
    job_label: User description for the beam job.
    runner: Beam Runner: (e.g. DataflowRunner, DirectRunner).
    project: GCP project ID (if DataflowRunner)
    region: GCP compute region (if DataflowRunner)
    output_dir: GCS or Local Path for output.
    compression: gzip or None.
    num_shards: Number of shards.
    schema_map: A schema map (Dictionary mapping Dataframe columns to types)
     used to derive the input and target schema.
    tfrecorder_wheel: Path to TFRecorder wheel for DataFlow
    dataflow_options: Dataflow Runner Options (optional)

  Returns:
    beam.Pipeline

  Note: These inputs must be validated upstream (by client.create_tfrecord())
  """

  job_name = _get_job_name(job_label)
  job_dir = _get_job_dir(output_dir, job_name)
  options = _get_pipeline_options(
      runner,
      job_name,
      job_dir,
      project,
      region,
      tfrecorder_wheel,
      dataflow_options)

  p = beam.Pipeline(options=options)
  with tft_beam.Context(temp_dir=os.path.join(job_dir, 'tft_tmp')):

    converter = schema.get_tft_coder(df.columns, schema_map)
    flatten_rows = ToCSVRows()

    # Each element in the data PCollection will be a dict
    # including the image_csv_columns and the image features created from
    # extract_images_fn.
    data = (
        p
        | 'ReadFromDataFrame' >> beam.Create(df.values.tolist())
        | 'ToCSVRows' >> beam.ParDo(flatten_rows)
        | 'DecodeCSV' >> beam.Map(converter.decode)
    )

    # Extract images if an image_uri key exists.
    image_uri_key = schema.get_key(schema.ImageUriType, schema_map)
    if image_uri_key:
      extract_images_fn = beam_image.ExtractImagesDoFn(image_uri_key)

      data = (
          data
          | 'ReadImage' >> beam.ParDo(extract_images_fn)
      )

    # If the schema contains a valid split key, partition the dataset.
    split_key = schema.get_key(schema.SplitKeyType, schema_map)

    # Note: This will not always reflect actual number of samples per dataset
    # written as TFRecords. The succeeding `Partition` operation may mark
    # additional samples from other splits as discarded. If a split has all
    # its samples discarded, the pipeline will still generate a TFRecord
    # file for that split, albeit empty.
    split_counts = get_split_counts(df, split_key)

    # Raw metadata is the TFT metadata after image insertion but before TFT
    # e.g Image columns have been added if necessary.
    raw_metadata = schema.get_raw_metadata(df.columns, schema_map)

    # Require training set to be available in the input data. The transform_fn
    # and transformed_metadata will be generated from the training set and
    # applied to the other datasets, if any
    assert 'TRAIN' in split_counts

    # Split dataset into train, validation, test sets.
    partition_fn = functools.partial(_partition_fn, split_key=split_key)
    train_data, val_data, test_data, discard_data = (
        data | 'SplitDataset' >> beam.Partition(
            partition_fn, len(schema.SplitKeyType.allowed_values)))

    raw_schema_map = schema.get_raw_schema_map(schema_map=schema_map)
    preprocessing_fn = functools.partial(
        _preprocessing_fn,
        schema_map=raw_schema_map)

    tfr_writer = functools.partial(
        _get_write_to_tfrecord, output_dir=job_dir, compress=compression,
        num_shards=num_shards)

    transform_fn = _transform_and_write_tfr(
        train_data, tfr_writer, preprocessing_fn=preprocessing_fn,
        raw_metadata=raw_metadata,
        label='Train')

    if 'VALIDATION' in split_counts:
      _transform_and_write_tfr(
          val_data, tfr_writer, transform_fn=transform_fn,
          raw_metadata=raw_metadata,
          label='Validation')

    if 'TEST' in split_counts:
      _transform_and_write_tfr(
          test_data, tfr_writer, transform_fn=transform_fn,
          raw_metadata=raw_metadata,
          label='Test')

    _ = (
        discard_data
        | 'WriteDiscardedData' >> beam.io.WriteToText(
            os.path.join(job_dir, 'discarded-data')))

    # Note: `transform_fn` already contains the transformed metadata
    _ = (transform_fn | 'WriteTransformFn' >> tft_beam.WriteTransformFn(
        job_dir))

  return p



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for TFRecorder types."""

import unittest
import tensorflow_transform as tft

from tfrecorder import schema


class SchemaTest(unittest.TestCase):
  """Tests for type module."""

  def test_valid_get_tft_coder(self):
    """Tests a valid call on get_tft_coder."""
    columns = ['split', 'image_uri', 'label']
    converter = schema.get_tft_coder(columns, schema.image_csv_schema)
    self.assertIsInstance(converter, tft.coders.CsvCoder)

  def test_valid_get_key(self):
    """Tests a valid split key."""
    key = schema.get_key(schema.SplitKeyType, schema.image_csv_schema)
    self.assertEqual(key, 'split')

  def test_no_get_split_key(self):
    """Tests no split key present."""
    test_schema = dict()
    for k, v in schema.image_csv_schema.items():
      # Brute force copy because OG is a FrozenOrderedDict.
      if k != 'split':
        test_schema[k] = v

    key = schema.get_key(schema.SplitKeyType, test_schema)
    self.assertIsNone(key)

  def test_get_raw_metadata(self):
    """Tests a valid call to get_raw_metadata."""
    columns = ['split', 'image_uri', 'label']
    raw_metadata = schema.get_raw_metadata(columns, schema.image_csv_schema)
    self.assertIsInstance(
        raw_metadata,
        tft.tf_metadata.dataset_metadata.DatasetMetadata)


if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for client."""

import os
from typing import List

import csv
import tempfile
import unittest

import mock
import pandas as pd

from tfrecorder import client
from tfrecorder import test_utils
from tfrecorder import schema


# pylint: disable=protected-access


class ClientTest(unittest.TestCase):
  """Misc tests for `client` module."""

  def setUp(self):
    self.test_df = test_utils.get_test_df()
    self.test_region = 'us-central1'
    self.test_project = 'foo'
    self.test_wheel = '/my/path/wheel.whl'

  @mock.patch('tfrecorder.client.beam_pipeline')
  def test_create_tfrecords_direct_runner(self, mock_beam):
    """Tests `create_tfrecords` Direct case."""
    mock_beam.build_pipeline().run().wait_until_finished.return_value = {
        'rows':6}
    r = client.create_tfrecords(
        self.test_df,
        runner='DirectRunner',
        output_dir='/tmp/direct_runner')
    self.assertTrue('metrics' in r)

  @mock.patch('tfrecorder.client.beam_pipeline')
  def test_create_tfrecords_dataflow_runner(self, mock_beam):
    """Tests `create_tfrecords` Dataflow case."""
    mock_beam.build_pipeline().run().job_id.return_value = 'foo_id'

    df2 = self.test_df.copy()
    df2['image_uri'] = 'gs://' + df2['image_uri']

    outdir = '/tmp/dataflow_runner'

    expected = {
        'job_id': 'foo_id',
        'dataflow_url': 'https://console.cloud.google.com/dataflow/jobs/' +
                        'us-central1/foo_id?project=foo'}

    os.makedirs(outdir, exist_ok=True)
    r = client.create_tfrecords(
        df2,
        runner='DataflowRunner',
        output_dir=outdir,
        region=self.test_region,
        project=self.test_project,
        tfrecorder_wheel=self.test_wheel)
    self.assertEqual(r, expected)

  def test_path_split(self):
    """Tests `_path_split`."""

    filename = 'image_file.jpg'
    dirpaths = ['/path/to/image/dir', 'gs://path/to/image/dir']
    for dir_ in dirpaths:
      filepath = os.path.join(dir_, filename)
      act_dirpath, act_filename = client._path_split(filepath)
      self.assertEqual(act_dirpath, dir_)
      self.assertEqual(act_filename, filename)


class InputValidationTest(unittest.TestCase):
  """'Tests for validation input data."""

  def setUp(self):
    self.test_df = test_utils.get_test_df()
    self.test_region = 'us-central1'
    self.test_project = 'foo'
    self.test_wheel = '/my/path/wheel.whl'
    self.test_schema_map = schema.image_csv_schema

  def test_valid_dataframe(self):
    """Tests valid DataFrame input."""
    self.assertIsNone(
        client._validate_data(
            self.test_df,
            schema.image_csv_schema))

  def test_missing_image(self):
    """Tests missing image column."""
    with self.assertRaises(AttributeError):
      df2 = self.test_df.copy()
      df2.drop('image_uri', inplace=True, axis=1)
      client._validate_data(df2, schema.image_csv_schema)

  def test_missing_label(self):
    """Tests missing label column."""
    with self.assertRaises(AttributeError):
      df2 = self.test_df.copy()
      df2.drop('label', inplace=True, axis=1)
      client._validate_data(df2, schema.image_csv_schema)

  def test_missing_split(self):
    """Tests missing split column."""
    with self.assertRaises(AttributeError):
      df2 = self.test_df.copy()
      df2.drop('split', inplace=True, axis=1)
      client._validate_data(df2, schema.image_csv_schema)

  def test_valid_runner(self):
    """Tests valid runner."""
    self.assertIsNone(client._validate_runner(
        runner='DirectRunner',
        project=self.test_project,
        region=self.test_region,
        tfrecorder_wheel=None))

  def test_invalid_runner(self):
    """Tests invalid runner."""
    with self.assertRaises(AttributeError):
      client._validate_runner(
          runner='FooRunner',
          project=self.test_project,
          region=self.test_region,
          tfrecorder_wheel=None)


  def test_gcs_path_with_dataflow_runner_missing_param(self):
    """Tests DataflowRunner with missing required parameter."""
    for p, r in [
        (None, self.test_region), (self.test_project, None), (None, None)]:
      with self.assertRaises(AttributeError) as context:
        client._validate_runner(
            runner='DataflowRunner',
            project=p,
            region=r,
            tfrecorder_wheel=self.test_wheel)
      self.assertTrue('DataflowRunner requires valid `project` and `region`'
                      in repr(context.exception))


  def test_gcs_path_with_dataflow_runner_missing_wheel(self):
    """Tests DataflowRunner with missing required whl path."""
    with self.assertRaises(AttributeError) as context:
      client._validate_runner(
          runner='DataflowRunner',
          project=self.test_project,
          region=self.test_region,
          tfrecorder_wheel=None)
      self.assertTrue('requires a tfrecorder whl file for remote execution.'
                      in repr(context.exception))


def _make_csv_tempfile(data: List[List[str]]) -> tempfile.NamedTemporaryFile:
  """Returns `NamedTemporaryFile` representing an image CSV."""

  f = tempfile.NamedTemporaryFile(mode='w+t', suffix='.csv')
  writer = csv.writer(f, delimiter=',')
  for row in data:
    writer.writerow(row)
  f.seek(0)
  return f


def get_sample_image_csv_data() -> List[List[str]]:
  """Returns sample CSV data in Image CSV format."""

  data = test_utils.get_test_data()
  header = list(data.keys())
  content = [list(row) for row in zip(*data.values())]
  return [header] + content


class ReadImageDirectoryTest(unittest.TestCase):
  """Tests `_read_image_directory`."""

  def setUp(self):
    self.image_data = test_utils.get_test_df()
    self.split_key = schema.get_key(
        schema.SplitKeyType, schema.image_csv_schema)
    self.label_key = schema.get_key(
        schema.StringLabelType, schema.image_csv_schema)
    self.image_uri_key = schema.get_key(
        schema.ImageUriType, schema.image_csv_schema)
    self.tempfiles = []
    self.tempdir = None

  def tearDown(self):
    for fp in self.tempfiles:
      fp.close()

    self.tempdir.cleanup()

  def test_normal(self):
    """Tests conversion of expected directory structure on local machine."""

    g = self.image_data.groupby([self.split_key, self.label_key])

    self.tempdir = tempfile.TemporaryDirectory()
    rows = []
    for (split, label), indices in g.groups.items():
      dir_ = os.path.join(self.tempdir.name, split, label)
      os.makedirs(dir_)
      for f in list(self.image_data.loc[indices, self.image_uri_key]):
        _, name = os.path.split(f)
        fp = tempfile.NamedTemporaryFile(
            dir=dir_, suffix='.jpg', prefix=name)
        self.tempfiles.append(fp)
        rows.append([split, fp.name, label])

    columns = list(schema.image_csv_schema.keys())
    actual = client._read_image_directory(self.tempdir.name)
    actual.sort_values(by=columns, inplace=True)
    actual.reset_index(drop=True, inplace=True)
    expected = pd.DataFrame(rows, columns=columns)
    expected.sort_values(by=columns, inplace=True)
    expected.reset_index(drop=True, inplace=True)
    pd.testing.assert_frame_equal(actual, expected)


class ReadCSVTest(unittest.TestCase):
  """Tests `read_csv`."""

  def setUp(self):
    data = get_sample_image_csv_data()
    self.header = data.pop(0)
    self.sample_data = data

  def test_valid_csv_no_header_no_names_specified(self):
    """Tests a valid CSV without a header and no header names given."""
    f = _make_csv_tempfile(self.sample_data)
    actual = client.read_csv(f.name, header=None)
    self.assertEqual(list(actual.columns), list(schema.image_csv_schema.keys()))
    self.assertEqual(actual.values.tolist(), self.sample_data)

  def test_valid_csv_no_header_names_specified(self):
    """Tests valid CSV without a header, but header names are given."""
    f = _make_csv_tempfile(self.sample_data)
    actual = client.read_csv(f.name, header=None, names=self.header)
    self.assertEqual(list(actual.columns), self.header)
    self.assertEqual(actual.values.tolist(), self.sample_data)

  def test_valid_csv_with_header_no_names_specified(self):
    """Tests valid CSV with header, and no header names given (inferred)."""

    f = _make_csv_tempfile([self.header] + self.sample_data)
    actual = client.read_csv(f.name)
    self.assertEqual(list(actual.columns), self.header)
    self.assertEqual(actual.values.tolist(), self.sample_data)

  def test_valid_csv_with_header_names_specified(self):
    """Tests valid CSV with header, and header names given (override)."""

    f = _make_csv_tempfile([self.header] + self.sample_data)
    actual = client.read_csv(f.name, names=self.header, header=0)
    self.assertEqual(list(actual.columns), self.header)
    self.assertEqual(actual.values.tolist(), self.sample_data)


class ToDataFrameTest(unittest.TestCase):
  """Tests `to_dataframe`."""

  def setUp(self) -> None:
    sample_data = get_sample_image_csv_data()
    columns = sample_data.pop(0)
    self.input_df = pd.DataFrame(sample_data, columns=columns)

  @mock.patch.object(client, 'read_csv', autospec=True)
  def test_input_csv(self, read_csv):
    """Tests valid input CSV file."""
    expected = self.input_df
    read_csv.return_value = expected
    f = _make_csv_tempfile(get_sample_image_csv_data())
    actual = client.to_dataframe(f.name)
    pd.testing.assert_frame_equal(actual, expected)

  def test_input_dataframe_no_names_specified(self):
    """Tests valid input dataframe with no header names specified."""
    actual = client.to_dataframe(self.input_df)
    pd.testing.assert_frame_equal(actual, self.input_df)

  def test_input_dataframe_with_header(self):
    """Tests valid input dataframe with header specified."""
    names = list(self.input_df.columns[0:-1])
    actual = client.to_dataframe(self.input_df, names=names)
    pd.testing.assert_frame_equal(actual, self.input_df[names])

  @mock.patch.object(client, '_read_image_directory')
  def test_input_image_dir(self, mock_fn):
    """Tests valid input image directory."""

    mock_fn.return_value = self.input_df

    with tempfile.TemporaryDirectory() as input_data:
      actual = client.to_dataframe(input_data)
      pd.testing.assert_frame_equal(actual, self.input_df)

  def test_error_invalid_inputs(self):
    """Tests error handling with different invalid inputs."""
    inputs = [0, 'not_a_csv_file', list(), dict()]
    for input_data in inputs:
      with self.assertRaises(ValueError):
        client.to_dataframe(input_data)


if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilites for writing tfrecorder tests."""

from typing import Any, Dict, List

import os

import numpy as np
from PIL import Image

from apache_beam.testing import test_pipeline
import pandas as pd

from tfrecorder import schema


TEST_DIR = 'tfrecorder/test_data'
TEST_TFRECORDS_DIR = os.path.join(TEST_DIR, 'sample_tfrecords')


def get_test_df() -> pd.DataFrame:
  """Gets a test dataframe that works with the data in test_data/."""
  return pd.read_csv(os.path.join(TEST_DIR, 'data.csv'))


def get_test_data() -> Dict[str, List[Any]]:
  """Returns test data in columnar format."""

  return get_test_df().to_dict(orient='list')


def get_raw_feature_df() -> pd.DataFrame:
  """Returns test dataframe having raw feature spec schema."""

  df = get_test_df()
  my_raw_schema = schema.get_raw_schema_map(schema.image_csv_schema)
  image_key = schema.get_key(schema.ImageUriType, schema.image_csv_schema)
  df.drop([image_key], axis=1, inplace=True)
  df['image_name'] = 'image_name'
  df['image'] = 'image'
  # Note: TF Transform parser expects string values in input. They will
  # be parsed based on the raw feature spec that is passed together with the
  # data
  df['image_height'] = '48'
  df['image_width'] = '48'
  df['image_channels'] = '3'
  df = df[my_raw_schema.keys()]

  return df


def get_test_pipeline():
  """Gets a test pipeline."""
  return test_pipeline.TestPipeline(runner='DirectRunner')


def make_random_image(height, width, channels):
  """Returns a random Numpy image."""

  return Image.fromarray(
      (np.random.random((height, width, channels)) * 255).astype(np.uint8))



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines input types for TFRecorder's input schema."""

import collections
from typing import Dict, List, Union

import frozendict
import tensorflow as tf
import tensorflow_transform as tft
from tensorflow_transform.tf_metadata import dataset_metadata
from tensorflow_transform.tf_metadata import schema_utils

# TODO(mikebernico): Refactor types into data classes
# All supported types will be based on SupportedType.
SupportedType = collections.namedtuple(
    'tfrecordInputType',
    ['type_name', 'feature_spec', 'allowed_values'])

# Supported type definitions here.
ImageUriType = SupportedType(
    type_name='image_uri',
    feature_spec=tf.io.FixedLenFeature([], tf.string),
    allowed_values=None)

# Note: split_key is an immutable type and these allowed values cannot change.
allowed_split_values = ['TRAIN', 'VALIDATION', 'TEST', 'DISCARD']
SplitKeyType = SupportedType(
    type_name='split_key',
    feature_spec=tf.io.FixedLenFeature([], tf.string),
    allowed_values=allowed_split_values)

#TODO(mikebernico): Implement in preprocess_fn
IntegerInputType = SupportedType(
    type_name='integer_input',
    feature_spec=tf.io.FixedLenFeature([], tf.int64),
    allowed_values=None)

#TODO(mikebernico): Implement in preprocess_fn
FloatInputType = SupportedType(
    type_name='float_input',
    feature_spec=tf.io.FixedLenFeature([], tf.float64),
    allowed_values=None)

#TODO(mikebernico): Implement in preprocess_fn
CategoricalInputType = SupportedType(
    type_name='categorical_input',
    feature_spec=tf.io.FixedLenFeature([], tf.string),
    allowed_values=None)

IntegerLabelType = SupportedType(
    type_name='integer_label',
    feature_spec=tf.io.FixedLenFeature([], tf.int64),
    allowed_values=None)

StringLabelType = SupportedType(
    type_name='string_label',
    feature_spec=tf.io.FixedLenFeature([], tf.string),
    allowed_values=None)

ImageSupportStringType = SupportedType(
    type_name='image_support_string',
    feature_spec=tf.io.FixedLenFeature([], tf.string),
    allowed_values=None)

ImageSupportIntType = SupportedType(
    type_name='image_support_int',
    feature_spec=tf.io.FixedLenFeature([], tf.int64),
    allowed_values=None)

# TODO(mikebernico): Refactor schema_map to a container class.
# Default schema supports the legacy image_csv format.
SchemaMap = Dict[str, SupportedType]

image_csv_schema = frozendict.FrozenOrderedDict({
    'split': SplitKeyType,
    'image_uri': ImageUriType,
    'label': StringLabelType})


def get_raw_schema_map(
    schema_map: Dict[str, collections.namedtuple]
    ) -> Dict[str, collections.namedtuple]:
  """Converts a schema to a raw (pre TFT / post image extraction) schema."""
  raw_schema = {}
  for k, v in schema_map.items():
    if v.type_name == 'image_uri':
      raw_schema['image_name'] = ImageSupportStringType
      raw_schema['image'] = ImageSupportStringType
      raw_schema['image_height'] = ImageSupportIntType
      raw_schema['image_width'] = ImageSupportIntType
      raw_schema['image_channels'] = ImageSupportIntType
    else:
      raw_schema[k] = schema_map[k]
  return raw_schema


def get_tft_coder(
    columns: List[str],
    schema_map: Dict[str, collections.namedtuple]
    ) -> tft.coders.CsvCoder:
  """Gets a TFT CSV Coder.

  Args:
    columns: Ordered DataFrame column names, from df.column.
    schema_map: Schema map used to infer the schema.

  Returns:
    tft.coders.CsvCoder
  """
  feature_spec = {}
  # Because the DF column name order may not match the feature_spec order
  # This maps existing column names to their feature spec (required part of
  # namedtuple)
  for col in columns:
    feature_spec[col] = schema_map[col].feature_spec

  metadata = dataset_metadata.DatasetMetadata(
      schema_utils.schema_from_feature_spec(feature_spec))

  return tft.coders.CsvCoder(columns, metadata.schema)


def get_key(
    type_: SupportedType,
    schema_map: Dict[str, collections.namedtuple]) -> Union[str, None]:
  """Gets first instance of key of type 'type_name' from schema map.

  Returns key name if present, otherwise returns None.
  """
  #TODO(mikebernico): Fix so that multiples of a key type work in future.
  for k, v in schema_map.items():
    if v.type_name == type_.type_name:
      return k
  return None


def get_raw_feature_spec(columns: List[str],
                         schema_map: Dict[str, collections.namedtuple]
                         ) -> Dict[str, tf.io.FixedLenFeature]:
  """Gets RAW (pre TFT) feature spec."""

  feature_spec = dict()

  # Because the DF column name order may not match the feature_spec order
  # this maps existing column names to their feature spec (req part of
  # namedtuple)
  for col in columns:
    if schema_map[col].type_name == 'image_uri':
      # Modify feature_spec for extracted image, don't include image_uri.
      # TODO(mikebernico) This only works in the case where the input has
      # ONLY 1 image. Generalize to multiple images someday?
      feature_spec['image_name'] = tf.io.FixedLenFeature([], tf.string)
      feature_spec['image'] = tf.io.FixedLenFeature([], tf.string)
      feature_spec['image_height'] = tf.io.FixedLenFeature([], tf.int64)
      feature_spec['image_width'] = tf.io.FixedLenFeature([], tf.int64)
      feature_spec['image_channels'] = tf.io.FixedLenFeature([], tf.int64)
    else:
      # Copy feature as-is.
      feature_spec[col] = schema_map[col].feature_spec
  return feature_spec


def get_raw_metadata(columns: List[str],
                     schema_map: Dict[str, collections.namedtuple]
                     ) -> dataset_metadata.DatasetMetadata:
  """Returns metadata prior to TF Transform preprocessing

  Note: takes base schema_map as input, not raw_schema_map.
  """
  feature_spec = get_raw_feature_spec(columns, schema_map)
  return dataset_metadata.DatasetMetadata(
      schema_utils.schema_from_feature_spec(feature_spec))



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a common interface for TFRecorder to DF Accessor and CLI.

client.py provides create_tfrecords() to upstream clients including
the Pandas DataFrame Accessor (accessor.py) and the CLI (cli.py).
"""
import collections
import logging
import os
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import apache_beam as beam
import pandas as pd
import tensorflow as tf

from tfrecorder import beam_pipeline
from tfrecorder import common
from tfrecorder import constants
from tfrecorder import schema

# TODO(mikebernico) Add test for only one split_key.
def _validate_data(df: pd.DataFrame,
                   schema_map: Dict[str, collections.namedtuple]):
  """Verifies data is consistent with schema."""

  for key, value in schema_map.items():
    _ = value # TODO(mikebernico) Implement type checking.
    if key not in df.columns:
      raise AttributeError(
          'DataFrame does not contain column {} listed in schema'.format(key))

def _validate_runner(
    runner: str,
    project: str,
    region: str,
    tfrecorder_wheel: str):
  """Validates an appropriate beam runner is chosen."""
  if runner not in ['DataflowRunner', 'DirectRunner']:
    raise AttributeError('Runner {} is not supported.'.format(runner))

  if (runner == 'DataflowRunner') & (
      any(not v for v in [project, region])):
    raise AttributeError(
        'DataflowRunner requires valid `project` and `region` to be specified.'
        'The `project` is {} and `region` is {}'.format(project, region))

  if (runner == 'DataflowRunner') & (not tfrecorder_wheel):
    raise AttributeError(
        'DataflowRunner requires a tfrecorder whl file for remote execution.')


def _path_split(filepath: str) -> Tuple[str, str]:
  """Splits `filepath` into (head, tail) where `tail` part after last '/'.

  e.g.
    filepath = '/path/to/image/file.jpg'
    head, tail = _path_split(filepath)
    # head -> '/path/to/image'
    # tail -> 'file.jpg'

  Similar to `os.path.split` but supports GCS paths (prefix: gs://).
  """

  if filepath.startswith(constants.GCS_PREFIX):
    _, path = filepath.split(constants.GCS_PREFIX)
    head, tail = os.path.split(path)
    return constants.GCS_PREFIX + head, tail

  return os.path.split(filepath)


def _read_image_directory(image_dir: str) -> pd.DataFrame:
  """Reads image data from a directory into a Pandas DataFrame.

  Expected directory structure:
    image_dir/
      <dataset split>/
        <label>/
          <image file>

  Example expected directory structure:
    image_dir/
      TRAIN/
        label0/
          image_000.jpg
          image_001.jpg
          ...
        label1/
          image_100.jpg
          ...
      VALIDATION/
        ...

  Output will be based on `schema.image_csv_schema`.
  The subdirectories should only contain image files.
  See `beam_image.load` for supported image formats.
  """

  rows = []
  split_values = schema.allowed_split_values
  for root, _, files in tf.io.gfile.walk(image_dir):
    if files:
      root_, label = _path_split(root)
      _, split = _path_split(root_)
      if split not in split_values:
        logging.warning('Unexpected split value: %s. Skipping %s',
                        split, root)
      # TODO(cezequiel): Add guard for non image files (e.g. .DS_Store)
      for f in files:
        image_uri = os.path.join(root, f)
        row = [split, image_uri, label]
        rows.append(row)

  return pd.DataFrame(rows, columns=schema.image_csv_schema.keys())


def _is_directory(input_data) -> bool:
  """Returns True if `input_data` is a directory; False otherwise."""

  return tf.io.gfile.isdir(input_data)


def read_csv(
    csv_file: str,
    header: Optional[Union[str, int, Sequence]] = 'infer',
    names: Optional[Sequence] = None) -> pd.DataFrame:
  """Returns a a Pandas DataFrame from a CSV file."""

  if header is None and not names:
    names = list(schema.image_csv_schema.keys())

  with tf.io.gfile.GFile(csv_file) as f:
    return pd.read_csv(f, names=names, header=header)


def to_dataframe(
    input_data: Union[str, pd.DataFrame],
    header: Optional[Union[str, int, Sequence]] = 'infer',
    names: Optional[Sequence] = None) -> pd.DataFrame:
  """Converts `input_data` to a Pandas DataFrame."""

  if isinstance(input_data, pd.DataFrame):
    df = input_data[names] if names else input_data

  elif isinstance(input_data, str) and input_data.endswith('.csv'):
    df = read_csv(input_data, header, names)

  elif isinstance(input_data, str) and _is_directory(input_data):
    df = _read_image_directory(input_data)

  else:
    raise ValueError('Unsupported `input_data`: {}'.format(type(input_data)))

  return df


def _get_beam_metric(
    metric_filter: beam.metrics.MetricsFilter,
    result: beam.runners.runner.PipelineResult,
    metric_type: str = 'counters') -> Optional[int]:
  """Queries a beam pipeline result for a specificed metric.

  Args:
    metric_filter: an instance of apache_beam.metrics.MetricsFilter()
    metric_type: A metric type (counters, distributions, etc.)

  Returns:
    Counter value or None
  """
  query_result = result.metrics().query(metric_filter)
  result_val = None
  if query_result[metric_type]:
    result_val = query_result[metric_type][0].result
  return result_val


def _configure_logging(logfile):
  """Configures logging options."""
  # Remove default handlers that TF set for us.
  logger = logging.getLogger('')
  logger.handlers = []
  handler = logging.FileHandler(logfile)
  logger.addHandler(handler)
  logger.setLevel(constants.LOGLEVEL)
  # This disables annoying Tensorflow and TFX info/warning messages on console.
  tf_logger = logging.getLogger('tensorflow')
  tf_logger.handlers = []
  tf_logger.addHandler(handler)

# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals

def create_tfrecords(
    source: Union[str, pd.DataFrame],
    output_dir: str,
    schema_map: Dict[str, collections.namedtuple] = schema.image_csv_schema,
    header: Optional[Union[str, int, Sequence]] = 'infer',
    names: Optional[Sequence] = None,
    runner: str = 'DirectRunner',
    project: Optional[str] = None,
    region: Optional[str] = None,
    tfrecorder_wheel: Optional[str] = None,
    dataflow_options: Optional[Dict[str, Any]] = None,
    job_label: str = 'create-tfrecords',
    compression: Optional[str] = 'gzip',
    num_shards: int = 0) -> Dict[str, Any]:
  """Generates TFRecord files from given input data.

  TFRecorder provides an easy interface to create image-based tensorflow records
  from a dataframe containing GCS locations of the images and labels.

  Usage:
    import tfrecorder

    job_id = tfrecorder.client.create_tfrecords(
        train_df,
        output_dir='gcs://foo/bar/train',
        runner='DirectFlowRunner)

  Args:
    source: Pandas DataFrame, CSV file or image directory path.
    output_dir: Local directory or GCS Location to save TFRecords to.
    schema_map: A dict mapping column names to supported types.
    header: Indicates row/s to use as a header. Not used when `input_data` is
      a Pandas DataFrame.
      If 'infer' (default), header is taken from the first line of a CSV
    runner: Beam runner. Can be 'DirectRunner' or 'DataFlowRunner'
    project: GCP project name (Required if DataflowRunner)
    region: GCP region name (Required if DataflowRunner)
    tfrecorder_wheel: Required for GCP Runs, path to the tfrecorder whl.
    dataflow_options: Options dict for DataflowRunner
    job_label: User supplied description for the Beam job name.
    compression: Can be 'gzip' or None for no compression.
    num_shards: Number of shards to divide the TFRecords into. Default is
        0 = no sharding.

  Returns:
    job_results: Dict
      job_id: Dataflow Job ID or 'DirectRunner'
      metrics: (optional) Beam metrics. Only used for DirectRunner
      dataflow_url: (optional) Job URL for DataflowRunner
  """

  df = to_dataframe(source, header, names)

  _validate_data(df, schema_map)
  _validate_runner(runner, project, region, tfrecorder_wheel)

  logfile = os.path.join('/tmp', constants.LOGFILE)
  _configure_logging(logfile)

  p = beam_pipeline.build_pipeline(
      df,
      job_label=job_label,
      runner=runner,
      project=project,
      region=region,
      output_dir=output_dir,
      compression=compression,
      num_shards=num_shards,
      schema_map=schema_map,
      tfrecorder_wheel=tfrecorder_wheel,
      dataflow_options=dataflow_options)

  result = p.run()

  if runner == 'DirectRunner':
    logging.info('Using DirectRunner - blocking until job completes.')
    result.wait_until_finish()

    row_count_filter = beam.metrics.MetricsFilter().with_name('row_count')
    good_image_filter = beam.metrics.MetricsFilter().with_name('image_good')
    bad_image_filter = beam.metrics.MetricsFilter().with_name('image_bad')

    row_count = _get_beam_metric(row_count_filter, result)
    good_image_count = _get_beam_metric(good_image_filter, result)
    bad_image_count = _get_beam_metric(bad_image_filter, result)

    # TODO(mikebernico): Profile metric impact with larger dataset.
    metrics = {
        'rows': row_count,
        'good_images': good_image_count,
        'bad_images': bad_image_count,
    }

    job_result = {
        'job_id': 'DirectRunner',
        'metrics': metrics
    }
    logging.info("Job Complete.")

  else:
    logging.info("Using Dataflow Runner.")
    # Construct Dataflow URL

    job_id = result.job_id()

    url = (
        constants.CONSOLE_DATAFLOW_URI +
        region +
        '/' +
        job_id +
        '?project=' +
        project)
    job_result = {
        'job_id': job_id,
        'dataflow_url': url
    }

  logging.shutdown()

  if runner == 'DataflowRunner':
    # if this is a Dataflow job, copy the logfile to GCS
    common.copy_logfile_to_gcs(logfile, output_dir)

  return job_result



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for beam_pipeline."""

import functools
import glob
import os
import tempfile
import unittest
from unittest import mock

import apache_beam as beam
import frozendict
import tensorflow as tf
from tensorflow_transform import beam as tft_beam

from tfrecorder import beam_pipeline
from tfrecorder import schema
from tfrecorder import test_utils


# pylint: disable=protected-access

class BeamPipelineTests(unittest.TestCase):
  """Tests for beam_pipeline.py"""

  def test_processing_fn_with_int_label(self):
    'Test preprocessing fn with integer label.'
    element = {
        'split': 'TRAIN',
        'image_uri': 'gs://foo/bar.jpg',
        'label': 1}
    my_schema = frozendict.FrozenOrderedDict({
        'split': schema.SplitKeyType,
        'image_uri': schema.ImageUriType,
        'label': schema.IntegerLabelType})

    result = beam_pipeline._preprocessing_fn(element, schema_map=my_schema)
    self.assertEqual(element, result)

  @mock.patch('tfrecorder.beam_pipeline.tft')
  def test_processing_fn_with_string_label(self, mock_transform):
    'Test preprocessing fn with string label.'
    mock_transform.compute_and_apply_vocabulary.return_value = tf.constant(
        0, dtype=tf.int64)
    element = {
        'split': 'TRAIN',
        'image_uri': 'gs://foo/bar.jpg',
        'label': tf.constant('cat', dtype=tf.string)}
    result = beam_pipeline._preprocessing_fn(element,
                                             schema_map=schema.image_csv_schema)
    result['label'] = result['label'].numpy()
    self.assertEqual(0, result['label'])

  def test_write_to_tfrecord(self):
    """Test _write_to_tfrecord() fn."""
    tfr_writer = beam_pipeline._get_write_to_tfrecord(
        output_dir='tmp',
        prefix='foo',
        compress=True,
        num_shards=2)
    self.assertIsInstance(tfr_writer, beam.io.tfrecordio.WriteToTFRecord)

  def test_partition_fn(self):
    """Test the partition function."""

    test_data = {
        'split': 'update_me',
        'image_uri': 'gs://foo/bar0.jpg',
        'label': 1}

    for i, part in enumerate(['TRAIN', 'VALIDATION', 'TEST', 'FOO']):
      test_data['split'] = part.encode('utf-8')
      index = beam_pipeline._partition_fn(test_data, split_key='split')

      self.assertEqual(
          index, i,
          '{} should be index {} but was index {}'.format(part, i, index))


class GetSplitCountsTest(unittest.TestCase):
  """Tests `get_split_counts` function."""

  def setUp(self):
    self.df = test_utils.get_test_df()
    self.schema_map = schema.image_csv_schema
    self.split_key = schema.get_key(schema.SplitKeyType, self.schema_map)

  def test_all_splits(self):
    """Tests case where train, validation and test data exists"""
    expected = {'TRAIN': 2, 'VALIDATION': 2, 'TEST': 2}
    actual = beam_pipeline.get_split_counts(self.df, self.split_key)
    self.assertEqual(actual, expected)

  def test_one_split(self):
    """Tests case where only one split (train) exists."""
    df = self.df[self.df.split == 'TRAIN']
    expected = {'TRAIN': 2}
    actual = beam_pipeline.get_split_counts(df, self.split_key)
    self.assertEqual(actual, expected)

  def test_error_no_split_key(self):
    """Tests case no split key/column exists."""
    df = self.df.drop(self.split_key, axis=1)
    with self.assertRaises(AssertionError):
      beam_pipeline.get_split_counts(df, self.split_key)

# pylint: disable=too-many-instance-attributes

class TransformAndWriteTfrTest(unittest.TestCase):
  """Tests `_transform_and_write_tfr` function."""

  def setUp(self):
    self.pipeline = test_utils.get_test_pipeline()
    self.raw_df = test_utils.get_raw_feature_df()
    self.temp_dir_obj = tempfile.TemporaryDirectory(dir='/tmp', prefix='test-')
    self.test_dir = self.temp_dir_obj.name
    self.tfr_writer = functools.partial(
        beam_pipeline._get_write_to_tfrecord, output_dir=self.test_dir,
        compress='gzip', num_shards=2)
    self.raw_schema = schema.get_raw_schema_map(schema.image_csv_schema)
    self.raw_metadata = schema.get_raw_metadata(self.raw_df.columns,
                                                self.raw_schema)
    self.converter = schema.get_tft_coder(self.raw_df.columns, self.raw_schema)
    self.transform_fn_path = ('./tfrecorder/test_data/sample_tfrecords')

  def tearDown(self):
    self.temp_dir_obj.cleanup()

  def _get_dataset(self, pipeline, df):
    """Returns dataset `PCollection`."""
    return (pipeline
            | beam.Create(df.values.tolist())
            | beam.ParDo(beam_pipeline.ToCSVRows())
            | beam.Map(self.converter.decode))

  def test_train(self):
    """Tests case where training data is passed."""

    with self.pipeline as p:
      with tft_beam.Context(temp_dir=os.path.join(self.test_dir, 'tmp')):
        df = self.raw_df[self.raw_df.split == 'TRAIN']
        dataset = self._get_dataset(p, df)
        preprocessing_fn = functools.partial(beam_pipeline._preprocessing_fn,
                                             schema_map=self.raw_schema)
        transform_fn = (
            beam_pipeline._transform_and_write_tfr(
                dataset, self.tfr_writer,
                preprocessing_fn=preprocessing_fn,
                raw_metadata=self.raw_metadata,
                label='Train'))
        _ = transform_fn | tft_beam.WriteTransformFn(self.test_dir)

    self.assertTrue(
        os.path.isdir(os.path.join(self.test_dir, 'transform_fn')))
    self.assertTrue(
        os.path.isdir(os.path.join(self.test_dir, 'transformed_metadata')))
    self.assertTrue(glob.glob(os.path.join(self.test_dir, 'train*.gz')))
    self.assertFalse(glob.glob(os.path.join(self.test_dir, 'validation*.gz')))
    self.assertFalse(glob.glob(os.path.join(self.test_dir, 'test*.gz')))

  def test_non_training(self):
    """Tests case where dataset contains non-training (e.g. test) data."""

    with self.pipeline as p:
      with tft_beam.Context(temp_dir=os.path.join(self.test_dir, 'tmp')):

        df = self.raw_df[self.raw_df.split == 'TEST']
        dataset = self._get_dataset(p, df)
        transform_fn = p | tft_beam.ReadTransformFn(self.transform_fn_path)
        beam_pipeline._transform_and_write_tfr(
            dataset, self.tfr_writer, transform_fn=transform_fn,
            raw_metadata=self.raw_metadata, label='Test')

    self.assertFalse(glob.glob(os.path.join(self.test_dir, 'train*.gz')))
    self.assertFalse(glob.glob(os.path.join(self.test_dir, 'validation*.gz')))
    self.assertTrue(glob.glob(os.path.join(self.test_dir, 'test*.gz')))


if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for beam_image."""

import base64
import unittest

import apache_beam as beam
from apache_beam.testing import util
import numpy as np
import PIL
from PIL import Image

from tfrecorder import beam_image
from tfrecorder import test_utils
from tfrecorder import schema


class BeamImageTests(unittest.TestCase):
  """Tests for beam_image.py"""

  def setUp(self):
    self.pipeline = test_utils.get_test_pipeline()
    self.df = test_utils.get_test_df()
    self.image_file = 'tfrecorder/test_data/images/cat/cat-640x853-1.jpg'

  def test_load(self):
    """Tests the image loading function."""
    img = beam_image.load(self.image_file)
    self.assertIsInstance(img, PIL.JpegImagePlugin.JpegImageFile)

  def test_file_not_found_load(self):
    """Test loading an image that doesn't exist."""
    with self.assertRaises(OSError):
      _ = beam_image.load('tfrecorder/test_data/images/cat/food.jpg')

  def test_mode_to_channel(self):
    """Tests `mode_to_channel`."""

    actual = [beam_image.mode_to_channel(mode)
              for mode in ('L', 'RGB', 'whatever')]
    self.assertEqual(actual, [1, 3, 3])

  def test_channel_to_mode(self):
    """Tests `channel_to_mode`."""

    actual = [beam_image.channel_to_mode(channel) for channel in (1, 2, 3)]
    self.assertEqual(actual, ['L', 'RGB', 'RGB'])

  def test_base64_encode(self):
    """Tests encode function."""
    img = beam_image.load(self.image_file)
    enc = beam_image.encode(img)
    decode = base64.b64decode(enc, altchars=beam_image.BASE64_ALTCHARS)
    self.assertEqual(img.tobytes(), decode)

  def test_base64_decode(self):
    """Tests `decode` function."""

    image = Image.open(self.image_file)
    b64_bytes = base64.b64encode(
        image.tobytes(), altchars=beam_image.BASE64_ALTCHARS)
    width, height = image.size
    actual = beam_image.decode(b64_bytes, width, height, 3)
    np.testing.assert_array_equal(np.asarray(actual), np.asarray(image))

  def test_extract_image_dofn(self):
    """Tests ExtractImageDoFn."""

    with self.pipeline as p:

      converter = schema.get_tft_coder(['split', 'image_uri', 'label'],
                                       schema.image_csv_schema)


      extract_images_fn = beam_image.ExtractImagesDoFn('image_uri')

      data = (
          p
          | 'ReadFromDataFrame' >> beam.Create(self.df.values.tolist())
          | 'FlattenDataFrame' >> beam.Map(
              lambda x: ','.join([str(item) for item in x]))
          | 'DecodeCSV' >> beam.Map(converter.decode)
          | 'ExtractImage' >> beam.ParDo(extract_images_fn)
      )

      def key_matcher(expected_keys):
        """Custom Beam Test Matcher that tests a list of keys exist."""
        def _equal(actual):
          """ _equal raises a BeamAssertException when an element in the
              PCollection doesn't contain the image extraction keys."""
          expected_keys_ = set(expected_keys)
          for element in actual:
            actual_keys = set(element.keys())
            if actual_keys != expected_keys_:
              raise util.BeamAssertException(
                  'PCollection key match failed. Actual ({}) vs. expected ({})'
                  .format(actual_keys, expected_keys_))
        return _equal

      expected_keys = ['image_name', 'label', 'split', 'image',
                       'image_height', 'image_width', 'image_channels']
      util.assert_that(data, key_matcher(expected_keys))



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFRecord Dataset utilities."""

import glob
import os
from typing import Dict

import tensorflow as tf
import tensorflow_transform as tft

from tfrecorder import schema


TRANSFORMED_METADATA_DIR = tft.TFTransformOutput.TRANSFORMED_METADATA_DIR
TRANSFORM_FN_DIR = tft.TFTransformOutput.TRANSFORM_FN_DIR

_FILE_EXT_TO_COMPRESSION_TYPE = {
    '.gz': 'GZIP',
    '.zlib': 'ZLIB',
}


def _validate_tfrecord_dir(tfrecord_dir: str):
  """Verifies that the TFRecord directory contains expected files."""

  # Check that TensorFlow Transform directories are present.
  for dirname in [TRANSFORMED_METADATA_DIR, TRANSFORM_FN_DIR]:
    if not os.path.isdir(os.path.join(tfrecord_dir, dirname)):
      raise FileNotFoundError(f'Missing expected directory: {dirname}')


# TODO(cezequiel): Add support for GCS files.
def _get_tfrecord_files_per_split(tfrecord_dir: str):
  """Returns TFRecord files for each split.

  The TFRecord filenames should have a prefix based on lowercase versions of
  items in `schema.allowed_split_values`. DISCARD split is not checked.
  """
  split_to_files = {}
  for split in schema.allowed_split_values[:-1]:
    prefix = split.lower()
    files = glob.glob(os.path.join(tfrecord_dir, prefix + '*'))
    if files:
      split_to_files[split] = files

  if not split_to_files:
    raise FileNotFoundError('No TFRecord files found.')

  return split_to_files


def _infer_tfrecord_compression_type(filename: str):
  """"Infers compression type (e.g. GZIP) of a TFRecord file.

  If file is uncompressed, returns ''.
  """

  ext = os.path.splitext(filename)[-1]
  return _FILE_EXT_TO_COMPRESSION_TYPE.get(ext, '')


def load(tfrecord_dir: str) -> Dict[str, tf.data.Dataset]:
  """Returns TF Datasets loaded from TFRecord files.

  This returns a `dict` keyed by dataset split, e.g.
    ```
    {
      'TRAIN': <tf.data.Dataset>,
      'VALIDATION': <tf.data.Dataset>,
      'TEST': <tf.data.Dataset>,
    }
    ```

  The `tfrecord_dir` is expected to have the following structure,
  based on TFRecorder's `create_tfrecords` default output:
    ```
    tfrecord_dir/
        train-*.tfrecord.gz
        validation-*.tfrecords.gz
        test-*.tfrecord.gz
        transformed_metadata/
        transform_fn/
    ```
  where the TFRecord file prefixes (e.g. train) would be have to match the
  excepted split values in lower case (see `schema.allowed_split_values`).

  This function will only generate items for splits that exists in the
  directory.
  """

  _validate_tfrecord_dir(tfrecord_dir)

  tft_output = tft.TFTransformOutput(tfrecord_dir)
  feature_spec = tft_output.transformed_feature_spec()
  splits = _get_tfrecord_files_per_split(tfrecord_dir)
  split_to_dataset = {}
  for key, files in splits.items():
    if not files:
      continue

    # Assumes all files in split have the same compression type
    first_file = files[0]
    compression_type = _infer_tfrecord_compression_type(first_file)

    dataset = tf.data.TFRecordDataset(files, compression_type)
    dataset = dataset.map(lambda x: tf.io.parse_single_example(x, feature_spec))
    split_to_dataset[key] = dataset

  return split_to_dataset



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Package setup."""

from setuptools import find_packages
from setuptools import setup


# Semantic versioning (PEP 440)
VERSION = '1.1.1'

REQUIRED_PACKAGES = [
    "apache-beam[gcp] >= 2.22.0",
    "avro >= 1.10.0",
    "coverage >= 5.1",
    "ipython >= 7.15.0",
    "fire >= 0.3.1",
    "frozendict >= 1.2",
    "nose >= 1.3.7",
    "numpy < 1.19.0",
    "pandas >= 1.0.4",
    "Pillow >= 7.1.2",
    "pyarrow >= 0.17, < 0.18.0",
    "pylint >= 2.5.3",
    "pytz >= 2020.1",
    "python-dateutil",
    "tensorflow == 2.3.1",
    "tensorflow_transform >= 0.22",
]


setup(
    name='tfrecorder',
    version=VERSION,
    install_requires=REQUIRED_PACKAGES,
    packages=find_packages(),
    include_package_data=True,
    description='TFRecorder creates TensorFlow Records easily.',
    entry_points={
        'console_scripts': ['tfrecorder=tfrecorder.cli:main'],
    },
)



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for checking content of TFRecord files."""

from typing import Dict, Optional, Sequence, Union

import csv
import os

import tensorflow as tf
import tensorflow_transform as tft

from tfrecorder import beam_image
from tfrecorder import common

_OUT_IMAGE_TEMPLATE = 'image_{:0>3d}.png'


def _stringify(scalar: tf.Tensor) -> str:
  """Converts scalar tensor into a Python string."""

  val = scalar.numpy()
  return val.decode('utf-8') if isinstance(val, bytes) else str(val)

# pylint: disable=too-many-locals
# TODO(cezequiel): deprecate in favor of `dataset.load`.
def _read_tfrecords(
    file_pattern: Union[str, Sequence[str]],
    tft_output_dir: Optional[str] = None,
    compression_type: str = 'GZIP') -> tf.data.Dataset:
  """Reads TFRecords files and outputs a TensorFlow Dataset.

  Currently supports Image CSV format only.
  """

  files = tf.io.gfile.glob(file_pattern)

  if not tft_output_dir:
    tft_output_dir = os.path.dirname(file_pattern)
  tf_transform_output = tft.TFTransformOutput(tft_output_dir)
  feature_spec = tf_transform_output.transformed_feature_spec()
  dataset = tf.data.TFRecordDataset(files, compression_type)
  return dataset.map(lambda x: tf.io.parse_single_example(
      x, feature_spec))


def _save_image_from_record(record: Dict[str, tf.Tensor], outfile: str):
  """Extracts image data from parsed TFRecord and saves it to a file."""

  b64_image = record['image'].numpy()
  image = beam_image.decode(
      b64_image,
      record['image_width'], record['image_height'], record['image_channels'])
  image.save(outfile)


def check_tfrecords(
    file_pattern: str,
    num_records: int = 1,
    output_dir: str = 'output',
    compression_type: str = 'GZIP'):
  """Reads TFRecord files and outputs decoded contents to a temp directory."""

  dataset = _read_tfrecords(file_pattern, compression_type=compression_type)

  data_dir = os.path.join(
      output_dir, 'check-tfrecords-' + common.get_timestamp())
  os.makedirs(data_dir)

  csv_file = os.path.join(data_dir, 'data.csv')
  with open(csv_file, 'wt') as f:
    writer = csv.writer(f)

    # Write CSV header
    for data in dataset.take(1).as_numpy_iterator():
      # .as_numpy_iterator() converts from Tuple of Tensors to a dict.
      # list() yields the keys of that dict.
      #TODO(mikebernico): Check the schema type instead of image key name.
      header = [k for k in list(data) if k != 'image']
      writer.writerow(header)

    for r in dataset.take(num_records):
      # Save non-image bytes data to CSV.
      # This will save image metadata as well.
      row = [_stringify(r[k]) for k in header]
      writer.writerow(row)

      # Save image data to a file
      if 'image_name' in r:
        _, image_filename = os.path.split(_stringify(r['image_name']))
        image_path = os.path.join(data_dir, image_filename)
        _save_image_from_record(r, image_path)

    print('Output written to {}'.format(data_dir))

    return data_dir
